# Cooperative Adaptive Cruise Control (CACC)

import sys
from AAPI import *
from PyANGKernel import *

# --------------------- inputs ----------------------
# section where the vehicles will be controlled
sectionId = 544
# maximum distance at which communication can be established
maximumDistance = 200.0 # m or ft, depending on the units set
# desired distance between a vehicle and its leader when CACC is active
desiredDistance = 10.0 # m
# controller gains for acceleration, velocity error and distance error
k1 = 1.0
k2 = 0.5
k3 = 0.1
# ---------------------------------------------------
model = GKSystem.getSystem().getActiveModel()
reactionTimes = dict()
reactionTimesAtStop = dict()
reactionTimesAtTrafficLight = dict()
sensitivityFactors = dict()
clearances = dict()
desiredSpeeds = dict()
caccVehs = list()

def AAPILoad():
	return 0

def AAPIInit():
	return 0

def AAPISimulationReady():
	#AKIPrintString( "AAPISimulationReady" )
	return 0

def AAPIManage(time, timeSta, timeTrans, acycle):
	return 0

# In this example finding the leader is trivial because we only have one section and one lane
# and the API returns the vehicles ordered from the end to the beginning of the section
# You should rewrite this function if you want to deal with more complex situations
# (multiple lanes, multiple sections)
def getLeader(vehPos):
	return vehPos - 1

def getSpeedMs(speed):
	if AKIInfNetGetUnits() == 1: # Metric units
		return speed / 3.6 # convert km/h -> m/s
	else: # English units
		return speed / 2.23694 # convert mph -> m/s
		
def getSpeedModelUnits(speed):
	if AKIInfNetGetUnits() == 1: # Metric units
		return speed * 3.6 # convert m/s -> km/h
	else: # English units
		return speed * 2.23694 # convert m/s -> mph

def getAccelerationMs2(acceleration):
	if AKIInfNetGetUnits() == 1: # Metric units
		return acceleration # it's already m/s2
	else: # English units
		return acceleration * 0.3048 # convert ft/s2 -> m/s2	

def getDistanceM(distance):
	if AKIInfNetGetUnits() == 1: # Metric units
		return distance # it's already m
	else: # English units
		return distance * 0.3048 # convert ft -> m

def CACC(egoState, leaderState, sectionId):	
	# Apply the CACC formula to calculate the desired acceleration
	leaderCurrentAcc = (getSpeedMs(leaderState.CurrentSpeed) - getSpeedMs(leaderState.PreviousSpeed)) / AKIGetSimulationStepTime() # m/s2
	errorSpd = getSpeedMs(leaderState.CurrentSpeed) - getSpeedMs(egoState.CurrentSpeed) # m/s
	gap = getDistanceM(((leaderState.xCurrentPosBack - egoState.xCurrentPos)**2 + (leaderState.yCurrentPosBack - egoState.yCurrentPos)**2)**0.5) # m
	errorDst = gap - desiredDistance
	caccAcc = (k1 * leaderCurrentAcc) + (k2 * errorSpd) + (k3 * errorDst) # m/s2
	
	# Constrain the calculated acceleration
	#acceleration = max(-1 * getAccelerationMs2(normalDecelerations[egoState.idVeh]), min(getAccelerationMs2(maxAccelerations[egoState.idVeh]), caccAcc)) # m/s2
	acceleration = max(-3.0, min(2.0, caccAcc)) # m/s2
	
	# Calculate the speed and constrain it
	newSpeed = min(desiredSpeeds[egoState.idVeh], getSpeedModelUnits(acceleration * AKIGetSimulationStepTime()) + egoState.CurrentSpeed) # km/h or mph
	
	print("Vehicle %s: leader's acceleration %s m/s2, speed error %s m/s, distance error %s m, CACC acceleration %s m/s2, constrained acceleration %s m/s2, current speed %s km/h or mph, new speed %s km/h or mph" % (egoState.idVeh, leaderCurrentAcc, errorSpd, errorDst, caccAcc, acceleration, egoState.CurrentSpeed, newSpeed))
	
	return newSpeed

def AAPIPostManage(time, timeSta, timeTrans, acycle):

	# For each vehicle in the section
	nbVehs = AKIVehStateGetNbVehiclesSection(sectionId, True)
	for vehPos in range(nbVehs):
	
		# Get the current state of the vehicle and that of its leader
		egoState = AKIVehStateGetVehicleInfSection(sectionId, vehPos)
		leaderState = AKIVehStateGetVehicleInfSection(sectionId, getLeader(vehPos))
		
		if egoState.report == 0:
			if leaderState.report == 0 and leaderState.CurrentPos - egoState.CurrentPos < maximumDistance:
				if egoState.idVeh not in caccVehs:
					print("Time %s: vehicle %s connected to vehicle %s" % (timeSta, egoState.idVeh, leaderState.idVeh))
					# Change behavioral parameters when CACC is activated
					parameters = AKIVehTrackedGetStaticInf(egoState.idVeh)
					parameters.reactionTime = AKIGetSimulationStepTime()
					parameters.reactionTimeAtStop = AKIGetSimulationStepTime()
					parameters.reactionTimeAtTrafficLight = AKIGetSimulationStepTime()
					parameters.sensitivityFactor = 0.0 # To avoid car-following constraining the speed
					parameters.minDistanceVeh = 0.0
					AKIVehTrackedSetStaticInf(egoState.idVeh, parameters)
					caccVehs.append(egoState.idVeh)
				
				# Calculate the speed by applying the CACC formula
				newSpeed = CACC(egoState, leaderState, sectionId)
		
				# Update the speed of the vehicle
				AKIVehTrackedModifySpeed(egoState.idVeh, newSpeed)
				
			else:
				if egoState.idVeh in caccVehs:
					print("Time %s: vehicle %s disconnected" % (timeSta, egoState.idVeh))
					# Restore behavioral parameters when CACC is deactivated
					parameters = AKIVehTrackedGetStaticInf(egoState.idVeh)
					parameters.reactionTime = reactionTimes[egoState.idVeh]
					parameters.reactionTimeAtStop = reactionTimesAtStop[egoState.idVeh]
					parameters.reactionTimeAtTrafficLight = reactionTimesAtTrafficLight[egoState.idVeh]
					parameters.sensitivityFactor = sensitivityFactors[egoState.idVeh]
					parameters.minDistanceVeh = clearances[egoState.idVeh]
					AKIVehTrackedSetStaticInf(egoState.idVeh, parameters)
					caccVehs.remove(egoState.idVeh)
	return 0

def AAPIEnterVehicle(idveh, idsection):
	return 0

def AAPIEnterVehicleSection(idveh, idsection, atime):
	# Set every vehicle that enters the section as tracked to be able to modify their speed afterwards
	if idsection == sectionId:
		AKIVehSetAsTracked(idveh)
		
		# Save vehicle parameters to alter them when activating CACC and restore them when deactivating CACC
		vehicleParameters = AKIVehTrackedGetStaticInf(idveh)
		reactionTimes[idveh] = vehicleParameters.reactionTime
		reactionTimesAtStop[idveh] = vehicleParameters.reactionTimeAtStop
		reactionTimesAtTrafficLight[idveh] = vehicleParameters.reactionTimeAtTrafficLight
		sensitivityFactors[idveh] = vehicleParameters.sensitivityFactor
		clearances[idveh] = vehicleParameters.minDistanceVeh # m or ft
		sectionParameters = AKIInfNetGetSectionANGInf(idsection)
		desiredSpeeds[idveh] = min(vehicleParameters.maxDesiredSpeed, vehicleParameters.speedAcceptance * sectionParameters.speedLimit) # km/h or mph
		
	return 0
	
def AAPIExitVehicle(idveh, idsection):
	return 0
	
def AAPIExitVehicleSection(idveh, idsection, atime):
	# Set every vehicle that exits the section as non-tracked
	if idsection == sectionId:
		AKIVehSetAsNoTracked(idveh)
		if idveh in caccVehs:
			caccVehs.remove(idveh)
		del reactionTimes[idveh]
		del reactionTimesAtStop[idveh]
		del reactionTimesAtTrafficLight[idveh]
		del sensitivityFactors[idveh]
		del clearances[idveh]
		del desiredSpeeds[idveh]
	return 0

def AAPIEnterPedestrian(idPedestrian, destinationCentroid):
	return 0

def AAPIExitPedestrian(idPedestrian, destinationCentroid):
	return 0

def AAPIFinish():
	return 0

def AAPIUnLoad():
	return 0

def AAPIPreRouteChoiceCalculation(time, timeSta):
	return 0

def AAPIVehicleStartParking (idveh, idsection, time):
	return 0